# Copyright (c) 2020 Mike Kinney

"""Unit tests for 'mech snapshot'."""
import re

from unittest.mock import patch
from pytest import raises

import mech.command
import mech.mech
import mech.vmrun


@patch('mech.utils.load_mechfile')
@patch('os.getcwd')
def test_mech_snapshot_list_no_mechdir(mock_os_getcwd, mock_load_mechfile, capfd,
                                       mechfile_two_entries):
    """Test 'mech snapshot list' with no '.mech' directory."""
    mock_load_mechfile.return_value = mechfile_two_entries
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        # root, dirs, files
        mock_walk.return_value = [('./tmp', [], []), ]
        arguments = {'<instance>': None}
        a_mech.list(arguments)
        mock_walk.assert_called()
        mock_load_mechfile.assert_called()
        out, _ = capfd.readouterr()
        # ensure a header prints out
        assert re.search(r'Snapshots', out, re.MULTILINE)


SNAPSHOT_LIST_WITHOUT_SNAPSHOTS = """Snapshots for instance:first
Total snapshots: 0
Snapshots for instance:second
Instance (second) is not created."""
@patch('mech.vmrun.VMrun.list_snapshots', return_value=SNAPSHOT_LIST_WITHOUT_SNAPSHOTS)
@patch('mech.utils.load_mechfile')
@patch('os.getcwd')
def test_mech_snapshot_list_no_snapshots(mock_os_getcwd, mock_load_mechfile,
                                         mock_list_snapshots, capfd,
                                         mechfile_two_entries):
    """Test 'mech snapshot list' without any snapshots."""
    mock_load_mechfile.return_value = mechfile_two_entries
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        mock_walk.return_value = [
            ('/tmp', ['first'], []),
            ('/tmp/first', [], ['some.vmx']),
        ]

        # with no args
        arguments = {'<instance>': None}
        a_mech.list(arguments)
        mock_walk.assert_called()
        mock_load_mechfile.assert_called()
        mock_list_snapshots.assert_called()
        out, _ = capfd.readouterr()
        assert re.search(r'Total snapshots: 0', out, re.MULTILINE)
        assert re.search(r'Instance \(second\) is not created.', out, re.MULTILINE)

        # single instance
        arguments = {'<instance>': 'first'}
        a_mech.list(arguments)
        out, _ = capfd.readouterr()
        mock_load_mechfile.assert_called()
        mock_list_snapshots.assert_called()
        assert re.search(r'Total snapshots: 0', out, re.MULTILINE)


SNAPSHOT_LIST_WITH_SNAPSHOT = """Snapshots for instance:first
Total snapshots: 1
snap1
Snapshots for instance:second
Instance (second) is not created."""
@patch('mech.vmrun.VMrun.list_snapshots', return_value=SNAPSHOT_LIST_WITH_SNAPSHOT)
@patch('mech.utils.load_mechfile')
@patch('os.getcwd')
def test_mech_snapshot_list_with_snapshot(mock_os_getcwd, mock_load_mechfile,
                                          mock_list_snapshots, capfd,
                                          mechfile_two_entries):
    """Test 'mech snapshot list' with a snapshots."""
    mock_load_mechfile.return_value = mechfile_two_entries
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        mock_walk.return_value = [
            ('/tmp', ['first'], []),
            ('/tmp/first', [], ['some.vmx']),
        ]

        # with no args
        arguments = {'<instance>': None}
        a_mech.list(arguments)
        mock_walk.assert_called()
        mock_load_mechfile.assert_called()
        mock_list_snapshots.assert_called()
        out, _ = capfd.readouterr()
        assert re.search(r'Total snapshots: 1', out, re.MULTILINE)
        assert re.search(r'Instance \(second\) is not created.', out, re.MULTILINE)

        # single instance
        arguments = {'<instance>': 'first'}
        a_mech.list(arguments)
        out, _ = capfd.readouterr()
        mock_load_mechfile.assert_called()
        mock_list_snapshots.assert_called()
        assert re.search(r'Total snapshots: 1', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.delete_snapshot')
@patch('mech.vmrun.VMrun.list_snapshots', return_value=SNAPSHOT_LIST_WITH_SNAPSHOT)
@patch('mech.utils.load_mechfile')
@patch('os.getcwd')
def test_mech_snapshot_delete_snapshot(mock_os_getcwd, mock_load_mechfile,
                                       mock_list_snapshots, mock_delete_snapshot, capfd,
                                       mechfile_two_entries):
    """Test 'mech snapshot delete'."""
    mock_load_mechfile.return_value = mechfile_two_entries
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        mock_walk.return_value = [
            ('/tmp', ['first'], []),
            ('/tmp/first', [], ['some.vmx']),
        ]

        arguments = {'<instance>': 'first'}
        a_mech.list(arguments)
        out, _ = capfd.readouterr()
        mock_load_mechfile.assert_called()
        mock_list_snapshots.assert_called()
        assert re.search(r'Total snapshots: 1', out, re.MULTILINE)

        arguments = {'<instance>': 'first', '<name>': 'snap2'}
        mock_delete_snapshot.return_value = None
        a_mech.delete(arguments)
        out, _ = capfd.readouterr()
        mock_delete_snapshot.assert_called()
        mock_list_snapshots.assert_called()
        assert re.search(r'Cannot delete', out, re.MULTILINE)

        arguments = {'<instance>': 'first', '<name>': 'snap1'}
        # Note: delete_snapshots return None if could not delete, or '' if it could
        mock_delete_snapshot.return_value = ''
        a_mech.delete(arguments)
        out, _ = capfd.readouterr()
        mock_delete_snapshot.assert_called()
        assert re.search(r' deleted', out, re.MULTILINE)

        arguments = {'<instance>': 'first'}
        mock_list_snapshots.return_value = SNAPSHOT_LIST_WITHOUT_SNAPSHOTS
        a_mech.list(arguments)
        out, _ = capfd.readouterr()
        mock_list_snapshots.assert_called()
        mock_delete_snapshot.assert_called()
        assert re.search(r'Total snapshots: 0', out, re.MULTILINE)


@patch('mech.utils.load_mechfile')
@patch('mech.utils.locate', return_value=None)
def test_mech_snapshot_list_not_created(mock_locate, mock_load_mechfile, capfd,
                                        mechfile_one_entry):
    """Test 'mech snapshot list' when vm is not created."""
    mock_load_mechfile.return_value = mechfile_one_entry
    global_arguments = {'--debug': False}
    arguments = {
        '<instance>': 'first',
    }
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    arguments = {'<instance>': 'first'}
    a_mech.list(arguments)
    out, _ = capfd.readouterr()
    mock_load_mechfile.assert_called()
    mock_locate.assert_called()
    assert re.search(r'not created', out, re.MULTILINE)


@patch('mech.utils.load_mechfile')
@patch('mech.utils.locate', return_value=None)
def test_mech_snapshot_save_not_created(mock_locate, mock_load_mechfile, capfd,
                                        mechfile_one_entry):
    """Test 'mech snapshot save' when vm is not created."""
    mock_load_mechfile.return_value = mechfile_one_entry
    global_arguments = {'--debug': False}
    arguments = {
        '<instance>': 'first',
    }
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    arguments = {'<instance>': 'first', '<name>': 'snap1'}
    a_mech.save(arguments)
    out, _ = capfd.readouterr()
    mock_locate.assert_called()
    mock_load_mechfile.assert_called()
    assert re.search(r'not created', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.snapshot', return_value='Snapshot (snap1) on VM (first) taken')
@patch('mech.utils.load_mechfile')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_mech_snapshot_save_success(mock_locate, mock_load_mechfile,
                                    mock_snapshot, capfd, mechfile_one_entry):
    """Test 'mech snapshot save' successful."""
    mock_load_mechfile.return_value = mechfile_one_entry
    global_arguments = {'--debug': False}
    arguments = {
        '<instance>': 'first',
    }
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    arguments = {'<instance>': 'first', '<name>': 'snap1'}
    a_mech.save(arguments)
    out, _ = capfd.readouterr()
    mock_locate.assert_called()
    mock_load_mechfile.assert_called()
    mock_snapshot.assert_called()
    assert re.search(r' taken', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.snapshot', return_value=None)
@patch('mech.utils.load_mechfile')
@patch('mech.utils.locate')
def test_mech_snapshot_save_failure(mock_locate, mock_load_mechfile,
                                    mock_vmrun_snapshot, mechfile_one_entry):
    """Test 'mech snapshot save' failure."""
    mock_locate.return_value = '/tmp/first/some.vmx'
    mock_load_mechfile.return_value = mechfile_one_entry
    global_arguments = {'--debug': False}
    arguments = {
        '<instance>': 'first',
    }
    a_mech = mech.mech.MechSnapshot(arguments=global_arguments)
    arguments = {'<instance>': 'first', '<name>': 'snap1'}
    with raises(SystemExit, match=r"Warning: Could not take snapshot."):
        a_mech.save(arguments)



# Copyright (c) 2020 Mike Kinney

"""Common pytest code."""
import json
import os
import pytest
import subprocess


from shutil import rmtree


@pytest.fixture
def mechfile_one_entry():
    """Return one mechfile entry."""
    return {
        'first': {
            'name': 'first',
            'box': 'bento/ubuntu-18.04',
            'box_version': '201912.04.0'
        }
    }


@pytest.fixture
def mechfile_one_entry_without_box_version():
    """Return one mechfile entry."""
    return {
        'first': {
            'name': 'first',
            'box': 'bento/ubuntu-18.04'
        }
    }


@pytest.fixture
def mechfile_one_entry_with_file():
    """Return one mechfile entry."""
    return {
        'name': 'first',
        'box': 'bento/ubuntu-18.04',
        'box_version': '201912.04.0',
        'file': '/tmp/somefile.box'
    }


@pytest.fixture
def mechfile_one_entry_with_auth():
    """Return one mechfile entry with auth."""
    return {
        'first': {
            'name': 'first',
            'box': 'bento/ubuntu-18.04',
            'box_version': '201912.04.0',
            'auth': {
                'username': 'bob',
                'pub_key': 'some_pub_key_data'
            }
        }
    }


@pytest.fixture
def mechfile_one_entry_with_auth_and_mech_use():
    """Return one mechfile entry with auth that has mech_use."""
    return {
        'first': {
            'name': 'first',
            'box': 'bento/ubuntu-18.04',
            'box_version': '201912.04.0',
            'auth': {
                'username': 'bob',
                'mech_use': 'true',
                'pub_key': 'some_pub_key_data'
            }
        }
    }


@pytest.fixture
def mechfile_two_entries():
    """Return two mechfile entries."""
    return {
        'first': {
            'name': 'first',
            'box': 'bento/ubuntu-18.04',
            'box_version': '201912.04.0',
            'shared_folders': [
                {
                    "host_path": ".",
                    "share_name": "mech"
                }
            ],
            'url':
            'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/'
            'versions/201912.04.0/providers/vmware_desktop.box'
        },
        'second': {
            'name': 'second',
            'box': 'bento/ubuntu-18.04',
            'box_version': '201912.04.0',
            'url':
            'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/'
            'versions/201912.04.0/providers/vmware_desktop.box'
        }
    }


CATALOG = """{
    "description": "foo",
    "short_description": "foo",
    "name": "bento/ubuntu-18.04",
    "versions": [
        {
            "version": "aaa",
            "status": "active",
            "description_html": "foo",
            "description_markdown": "foo",
            "providers": [
                {
                    "name": "vmware_desktop",
                    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/aaa/providers/vmware_desktop.box",
                    "checksum": null,
                    "checksum_type": null
                }
            ]
        }
    ]
}"""
@pytest.fixture
def catalog():
    """Return a catalog."""
    return CATALOG


@pytest.fixture
def shell_provision_config():
    return [
        {
            "type": "shell",
            "path": "file1.sh",
            "args": [
                "a=1",
                "b=true",
            ],
        },
        {
            "type": "shell",
            "inline": "echo hello from inline"
        },
        {
            "type": "shell",
            "inline": "echo hello2 from inline",
            "args": []
        }
    ]


@pytest.fixture
def pyinfra_provision_config():
    return [
        {
            "type": "pyinfra",
            "path": "file1.py",
            "args": [
                "sudo=True",
                "b=1",
            ],
        },
        {
            "type": "pyinfra",
            "path": "file2.py",
        }
    ]


@pytest.fixture
def pyinfra_provision_http_config():
    return [
        {
            "type": "pyinfra",
            "path": "https://github.com/Fizzadar/pyinfra/blob/"
            "master/examples/docker_ce.py",
            "args": [
                "sudo=True"
            ],
        }
    ]


@pytest.fixture
def catalog_as_json():
    """Return a catalog as json."""
    return json.loads(CATALOG)


@pytest.fixture
def mech_add_arguments():
    """Return the default 'mech add' arguments."""
    return {
        '--force': False,
        '--box-version': None,
        '--name': None,
        '--box': None,
        '--add-me': None,
        '--use-me': None,
        '--provider': None,
        '<location>': None,
    }


@pytest.fixture
def mech_box_arguments():
    """Return the default 'mech box' arguments."""
    return {
        '--force': False,
        '--box-version': None,
        '<provider>': None,
        '<location>': None,
    }


@pytest.fixture
def mech_init_arguments():
    """Return the default 'mech init' arguments."""
    return {
        '--force': False,
        '--box-version': None,
        '--name': None,
        '--box': None,
        '--add-me': None,
        '--use-me': None,
        '--provider': None,
        '<location>': None,
    }


class Helpers:
    @staticmethod
    def get_mock_data_written(a_mock):
        """Helper function to get the data written to a mocked file."""
        written = ''
        for call in a_mock.mock_calls:
            tmp = '{}'.format(call)
            if tmp.startswith('call().write('):
                line = tmp.replace("call().write('", '')
                line = line.replace("')", '')
                line = line.replace("\\n", '\n')
                written += line
        return written

    @staticmethod
    def kill_pids(pids):
        """Kill all pids."""
        for pid in pids:
            results = subprocess.run(args='kill {}'.format(pid), shell=True, capture_output=True)
            if results.returncode != 0:
                print("Could not kill pid:{}".format(pid))

    @staticmethod
    def find_pids(search_string):
        """Return all pids that that match the search_string."""
        # print('search_string:{}'.format(search_string))
        pids = []
        results = subprocess.run(args="ps -ef | grep '{}' | grep -v grep"
                                 .format(search_string), shell=True, capture_output=True)
        # print('results:{}'.format(results))
        if results.returncode == 0:
            # we found a proc
            stdout = results.stdout.decode('utf-8')
            for line in stdout.split('\n'):
                data = line.split()
                if len(data) > 2:
                    # add pid to the collection
                    pids.append(data[1])
        # print('pids:{}'.format(pids))
        return pids

    @staticmethod
    def cleanup_dir_and_vms_from_dir(a_dir, names=['first']):
        """Kill processes and remove directory and re-create the directory.
           For VMware VMs, look for the .vmx part_of_dir matches the full path.
           For virtualbox look 'VBoxHeadless --comment <name of the instance>'
           Note: Unfortunately, the name is global, so you can only have one
           'first' isntance. Make instance name unique for int tests that use
           virtualbox.
       """
        # remove from virtualbox
        for name in names:
            results = subprocess.run(args="VBoxManage unregistervm {}".format(name),
                                     shell=True, capture_output=True)
            # print('results:{}'.format(results))
        # kill vmware processes
        Helpers.kill_pids(Helpers.find_pids('vmware-vmx.*' + a_dir + '/.mech/'))
        # kill virtualbox processes (if any)
        for name in names:
            Helpers.kill_pids(Helpers.find_pids('VBoxHeadless --comment {} '.format(name)))
        # clean up the vm files
        rmtree(a_dir, ignore_errors=True)
        # remove any "dead" instances in virtualbox
        results = subprocess.run(args="VBoxManage list vms", shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        vms = stdout.split('\n')
        # print('vms:{}'.format(vms))
        for line in vms:
            # print('line:{}'.format(line))
            parts = line.split(' ')
            if len(parts) > 1 and parts[0] == '"<inaccessible>"':
                results = subprocess.run(args="VBoxManage unregistervm {}".format(parts[1]),
                                         shell=True, capture_output=True)
                # print('results:{}'.format(results))
        # print('results:{}'.format(results))
        # re-create the directory to start afresh
        os.mkdir(a_dir)


@pytest.fixture
def helpers():
    """Helper functions for testing."""
    return Helpers



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""Handle the mech options using docopt."""

from __future__ import absolute_import

import re
import textwrap

import docopt

from .compat import get_meth_func


NBSP = '__'


def cmd_usage(doc):
    """Show the command usage."""
    return doc.replace(NBSP, ' ')


docopt_extras_ref = docopt.extras


def docopt_extras(the_help, version, options, doc):
    """Show the "Extra" help info."""
    return docopt_extras_ref(the_help, version, options, cmd_usage(doc))


def DocoptExit____init__(self, message=''):
    """Constructor for docopt."""
    SystemExit.__init__(self, (message + '\n' + cmd_usage(self.usage)).strip())


docopt.extras = docopt_extras
docopt.DocoptExit.__init__ = DocoptExit____init__


def spaced(name):
    """Return the command name."""
    name = re.sub(r'[ _]+', r' ', name)
    name = re.sub(r'(?<=[^_])([A-Z])', r' \1', name).lower()
    return re.sub(r'^( *)(.*?)( *)$', r'\2', name)


class Command():
    """
    Usage: command <subcommand> [<args>...]
    """

    subcommand_name = '<subcommand>'
    argv_name = '<args>'

    @staticmethod
    def docopt(doc, **kwargs):
        """Parse comments for arguments."""
        name = kwargs.pop('name', "")
        name = spaced(name)
        doc = textwrap.dedent(doc).replace(name, name.replace(' ', NBSP))
        arguments = docopt.docopt(doc, options_first=True, **kwargs)
        return arguments

    def __init__(self, arguments):
        self.arguments = arguments

    def __call__(self):
        """Invoke the command with the arguments."""
        if self.subcommand_name in self.arguments:
            cmd = self.arguments[self.subcommand_name]
            cmd_attr = cmd.replace('-', '_')
            if hasattr(self, cmd_attr):
                klass = getattr(self, cmd_attr)
                meth_func = get_meth_func(klass)
                if meth_func:
                    cmd = meth_func.__name__.replace('_', '-')
                name = '{} {}'.format(self.__class__.__name__, cmd)
                if klass.__doc__:
                    arguments = self.docopt(klass.__doc__,
                                            argv=self.arguments.get(self.argv_name, []), name=name)
                else:
                    arguments = []
                obj = klass(arguments)
            else:
                obj = self.run()
        else:
            obj = self.run()
        if callable(obj):
            obj = obj()
        return obj

    def run(self):
        """Run the command."""
        raise docopt.DocoptExit()



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: provisioning tests"""
import re
import subprocess

import pytest


@pytest.mark.int
def test_int_provision(helpers):
    """Provision testing."""

    test_dir = "tests/int/provision/tmp"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    # copy files from parent dir
    command = "cp ../file* .; cp ../Mechfile ."
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # up without provisioning
    command = "mech up --disable-provisioning"
    expected_lines = [r".first.*started",
                      r".second.*started",
                      r".third.*started",
                      r".fourth.*started"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # show provisioning
    command = "mech provision -s"
    expected_lines = [r"first.*Provision 2 entries",
                      r"second.*Provision 3 entries",
                      r"fourth.*Provision 1 entries",
                      r"Nothing"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure there is no file on first
    command = 'mech ssh -c "ls -al /tmp/file1.txt" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("No such file or directory", stdout)
    assert results.returncode == 1

    # ensure there is no file on second (testing shell non-inline shell provisioning)
    command = 'mech ssh -c "ls -al /tmp/file1.sh.out" second'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("No such file or directory", stdout)
    assert results.returncode == 1

    # ensure there is no file on second (testing shell inline shell provisioning)
    command = 'mech ssh -c "ls -al /tmp/inline_test.out" second'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("No such file or directory", stdout)
    assert results.returncode == 1

    # ensure the package 'asterisk' is not installed on fourth
    command = 'mech ssh -c "apk list asterisk" fourth'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("installed", stdout) is None
    assert results.returncode == 0

    # provision
    command = "mech provision"
    expected_lines = [r"first.*Provision 2 entries",
                      r"second.*Provision 3 entries",
                      r"fourth.*Provision 1 entries",
                      r"Nothing"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure file exists now on first
    command = 'mech ssh -c "ls -al /tmp/file1.txt" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("/tmp/file1.txt", stdout)
    assert results.returncode == 0

    # ensure file exists now on second from non-inline shell provisioning
    command = 'mech ssh -c "ls -al /tmp/file1.sh.out" second'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("/tmp/file1.sh.out", stdout)
    assert results.returncode == 0

    # ensure file1.sh.out has the args
    command = """mech ssh -c "grep 'a=1 b=true' /tmp/file1.sh.out" second"""
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("a=1 b=true", stdout)
    assert results.returncode == 0

    # ensure file exists now on second from inline shell provisioning
    command = 'mech ssh -c "ls -al /tmp/inline_test.out" second'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("/tmp/inline_test.out", stdout)
    assert results.returncode == 0

    # ensure package 'asterisk' is installed on fourth (should have been installed
    # as part of the pyinfra provisioining)
    command = 'mech ssh -c "apk list asterisk" fourth'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("installed", stdout)
    assert results.returncode == 0

    # ensure provisioning runs during "up"
    command = "mech up"
    expected_lines = [r"first.*Provision 2 entries",
                      r"second.*Provision 3 entries",
                      r"fourth.*Provision 1 entries",
                      r"Nothing"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up at the end
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""MechSnapshot class"""

from __future__ import print_function, absolute_import

import sys
import logging

from clint.textui import colored

from .vmrun import VMrun
from .mech_instance import MechInstance
from .mech_command import MechCommand

LOGGER = logging.getLogger(__name__)


class MechSnapshot(MechCommand):
    """
    Usage: mech snapshot <subcommand> [<args>...]

    Available subcommands:
        (delete|remove)   delete a snapshot taken previously with snapshot save
        (list|ls)         list all snapshots taken for a machine
        save              take a snapshot of the current state of the machine

    For help on any individual subcommand run `mech snapshot <subcommand> -h`
    """

    def delete(self, arguments):  # pylint: disable=no-self-use
        """
        Delete a snapshot taken previously with snapshot save.

        Usage: mech snapshot delete [options] <name> <instance>

        Options:
            -h, --help                       Print this help
        """
        name = arguments['<name>']

        instance = arguments['<instance>']
        inst = MechInstance(instance)

        if inst.provider == 'vmware':
            vmrun = VMrun(inst.vmx)
            if vmrun.delete_snapshot(name) is None:
                print(colored.red("Cannot delete name"))
            else:
                print(colored.green("Snapshot {} deleted".format(name)))
        else:
            print(colored.red("Not yet implemented on this platform."))

    # add alias for 'mech snapshot remove'
    remove = delete

    def list(self, arguments):
        """
        List all snapshots taken for a machine.

        Usage: mech snapshot list [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)
            print('Snapshots for instance:{}'.format(instance))
            if inst.created:
                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    print(vmrun.list_snapshots())
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print(colored.red('Instance ({}) is not created.'.format(instance)))

    # add alias for 'mech snapshot ls'
    ls = list

    def save(self, arguments):  # pylint: disable=no-self-use
        """
        Take a snapshot of the current state of the machine.

        Usage: mech snapshot save [options] <name> <instance>

        Notes:
            Take a snapshot of the current state of the machine.

            Snapshots are useful for experimenting in a machine and being able
            to rollback quickly.

        Options:
            -h, --help                       Print this help
        """
        name = arguments['<name>']
        instance = arguments['<instance>']

        inst = MechInstance(instance)
        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                if vmrun.snapshot(name) is None:
                    sys.exit(colored.red("Warning: Could not take snapshot."))
                else:
                    print(colored.green("Snapshot ({}) on VM ({}) taken".format(name, instance)))
            else:
                print(colored.red("Not yet implemented on this platform."))
        else:
            print(colored.red('Instance ({}) is not created.'.format(instance)))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""MechInstance class"""

from __future__ import print_function, absolute_import

import os
import re
import sys
import logging
import textwrap

from clint.textui import colored

from . import utils
from .vmrun import VMrun
from .vbm import VBoxManage

LOGGER = logging.getLogger(__name__)

DEFAULT_USER = 'vagrant'
DEFAULT_PASSWORD = 'vagrant'
INSECURE_PRIVATE_KEY = """-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEA6NF8iallvQVp22WDkTkyrtvp9eWW6A8YVr+kz4TjGYe7gHzI
w+niNltGEFHzD8+v1I2YJ6oXevct1YeS0o9HZyN1Q9qgCgzUFtdOKLv6IedplqoP
kcmF0aYet2PkEDo3MlTBckFXPITAMzF8dJSIFo9D8HfdOV0IAdx4O7PtixWKn5y2
hMNG0zQPyUecp4pzC6kivAIhyfHilFR61RGL+GPXQ2MWZWFYbAGjyiYJnAmCP3NO
Td0jMZEnDkbUvxhMmBYSdETk1rRgm+R4LOzFUGaHqHDLKLX+FIPKcF96hrucXzcW
yLbIbEgE98OHlnVYCzRdK8jlqm8tehUc9c9WhQIBIwKCAQEA4iqWPJXtzZA68mKd
ELs4jJsdyky+ewdZeNds5tjcnHU5zUYE25K+ffJED9qUWICcLZDc81TGWjHyAqD1
Bw7XpgUwFgeUJwUlzQurAv+/ySnxiwuaGJfhFM1CaQHzfXphgVml+fZUvnJUTvzf
TK2Lg6EdbUE9TarUlBf/xPfuEhMSlIE5keb/Zz3/LUlRg8yDqz5w+QWVJ4utnKnK
iqwZN0mwpwU7YSyJhlT4YV1F3n4YjLswM5wJs2oqm0jssQu/BT0tyEXNDYBLEF4A
sClaWuSJ2kjq7KhrrYXzagqhnSei9ODYFShJu8UWVec3Ihb5ZXlzO6vdNQ1J9Xsf
4m+2ywKBgQD6qFxx/Rv9CNN96l/4rb14HKirC2o/orApiHmHDsURs5rUKDx0f9iP
cXN7S1uePXuJRK/5hsubaOCx3Owd2u9gD6Oq0CsMkE4CUSiJcYrMANtx54cGH7Rk
EjFZxK8xAv1ldELEyxrFqkbE4BKd8QOt414qjvTGyAK+OLD3M2QdCQKBgQDtx8pN
CAxR7yhHbIWT1AH66+XWN8bXq7l3RO/ukeaci98JfkbkxURZhtxV/HHuvUhnPLdX
3TwygPBYZFNo4pzVEhzWoTtnEtrFueKxyc3+LjZpuo+mBlQ6ORtfgkr9gBVphXZG
YEzkCD3lVdl8L4cw9BVpKrJCs1c5taGjDgdInQKBgHm/fVvv96bJxc9x1tffXAcj
3OVdUN0UgXNCSaf/3A/phbeBQe9xS+3mpc4r6qvx+iy69mNBeNZ0xOitIjpjBo2+
dBEjSBwLk5q5tJqHmy/jKMJL4n9ROlx93XS+njxgibTvU6Fp9w+NOFD/HvxB3Tcz
6+jJF85D5BNAG3DBMKBjAoGBAOAxZvgsKN+JuENXsST7F89Tck2iTcQIT8g5rwWC
P9Vt74yboe2kDT531w8+egz7nAmRBKNM751U/95P9t88EDacDI/Z2OwnuFQHCPDF
llYOUI+SpLJ6/vURRbHSnnn8a/XG+nzedGH5JGqEJNQsz+xT2axM0/W/CRknmGaJ
kda/AoGANWrLCz708y7VYgAtW2Uf1DPOIYMdvo6fxIB5i9ZfISgcJ/bbCUkFrhoH
+vq/5CIWxCPp0f85R4qxxQ5ihxJ0YDQT9Jpx4TMss4PSavPaBH3RXow5Ohe+bYoQ
NE5OgEXk2wVfZczCZpigBKbKZHNYcelXtTt/nP3rsCuGcM4h53s=
-----END RSA PRIVATE KEY-----
"""


class MechInstance():
    """Class to hold a mech instance (aka virtual machine)."""

    def __init__(self, name, mechfile=None):
        """Constructor for the mech instance."""
        if not name or name == "":
            raise AttributeError("Must provide a name for the instance.")
        if not mechfile:
            mechfile = utils.load_mechfile()
        LOGGER.debug("loaded mechfile:%s", mechfile)
        if mechfile.get(name, None):
            self.name = name
        else:
            sys.exit(colored.red("Instance ({}) was not found in the "
                                 "Mechfile".format(name)))
        self.box = mechfile[name].get('box', None)
        self.box_version = mechfile[name].get('box_version', None)
        self.url = mechfile[name].get('url', None)
        self.box_file = mechfile[name].get('file', None)
        self.provision = mechfile[name].get('provision', None)
        self.enable_ip_lookup = False
        self.config = {}
        self.ip = None
        # Note: providers are 'vmware' (default) or 'virtualbox'.
        self.provider = mechfile[name].get('provider', 'vmware')
        self.tools_state = None
        self.auth = mechfile[name].get('auth', None)
        self.shared_folders = mechfile[name].get('shared_folders', [])
        self.user = DEFAULT_USER
        self.password = DEFAULT_PASSWORD
        self.use_psk = False
        self.path = os.path.join(utils.mech_dir(), name)
        self.vmx = None
        self.vbox = None
        if self.provider == 'vmware':
            # Note: If vmware vm has not been started vmx will be None
            vmx = utils.locate(self.path, '*.vmx')
            if vmx:
                self.vmx = vmx
                self.created = True
            else:
                self.vmx = None
                self.created = False
        elif self.provider == 'virtualbox':
            # Note: If virtualbox vm has not been started vmx will be None
            vbox = utils.locate(self.path, '*.vbox')
            if vbox:
                self.vbox = vbox
                self.created = True
            else:
                self.vbox = None
                self.created = False

        self.no_nat = False
        self.remove_vagrant = False
        self.gui = False
        self.disable_provisioning = False
        self.disable_shared_folders = False

        # If vmx exists, then the VM has already been created.
        # See if we need to switch to preshared key authentication
        # for interactions with this guest.
        if self.created:
            self.switch_to_psk()

    def switch_to_psk(self):
        """Switch to using preshared key, instead of using user/password."""
        if self.auth:
            mech_use = self.auth.get('mech_use', False)
            username = self.auth.get('username')
            if username and username != '' and mech_use:
                self.user = username
                self.password = None
                self.use_psk = True

    def get_ip(self, wait=False, quiet=True):
        """ Get the ip address."""
        if self.ip:
            return self.ip
        else:
            if self.provider == 'vmware':
                if self.vmx:
                    vmrun = VMrun(self.vmx)
                    ip_address = vmrun.get_guest_ip_address(wait=wait,
                                                            lookup=self.enable_ip_lookup,
                                                            quiet=quiet)
                    self.ip = ip_address
                    return self.ip
            else:
                vbm = VBoxManage()
                self.ip = vbm.ip(self.name, wait=wait, quiet=quiet)
                return self.ip

    def get_vm_state(self):
        """ Get the state of the VM.
            Returns info like: ('running', 'paused', 'powered off')
        """
        if self.provider == 'vmware':
            vmrun = VMrun(self.vmx)
            return vmrun.vm_state()

        if self.provider == 'virtualbox':
            vbm = VBoxManage()
            return vbm.vm_state(self.name)

    def get_vm_info(self):
        """ Get detailed info about the VM.
            There is no equivalent in VMware. :-(
        """
        if self.provider == 'virtualbox':
            vbm = VBoxManage()
            return vbm.get_vm_info(self.name)

    def get_tools_state(self):
        """ Get the tools state."""
        if self.tools_state:
            return self.tools_state
        else:
            if self.vmx:
                vmrun = VMrun(self.vmx)
                self.tools_state = vmrun.installed_tools()
                return self.tools_state

    def __repr__(self):
        """Return a representation of a Mech instance."""
        sep = '\n'
        return ('name:{name}{sep}created:{created}{sep}box:{box}'
                '{sep}box_version:{box_version}'
                '{sep}url:{url}{sep}box_file:{box_file}{sep}provision:{provision}'
                '{sep}vmx:{vmx}{sep}user:{user}'
                '{sep}password:{password}{sep}enable_ip_lookup:{enable_ip_lookup}'
                '{sep}config:{config}{sep}shared_folders:{shared_folders}'
                '{sep}auth:{auth}{sep}use_psk:{use_psk}{sep}provider:{provider}'
                '{sep}no_nat:{no_nat}{sep}remove_vagrant:{remove_vagrant}'
                '{sep}gui:{gui}{sep}disable_provisioning:{disable_provisioning}'
                '{sep}disable_shared_folders:{disable_shared_folders}'.
                format(name=self.name, created=self.created,
                       box=self.box, box_version=self.box_version,
                       url=self.url, box_file=self.box_file,
                       provision=self.provision, vmx=self.vmx,
                       user=self.user, password=self.password,
                       enable_ip_lookup=self.enable_ip_lookup,
                       config=self.config,
                       shared_folders=self.shared_folders,
                       auth=self.auth, use_psk=self.use_psk,
                       provider=self.provider, no_nat=self.no_nat,
                       remove_vagrant=self.remove_vagrant, gui=self.gui,
                       disable_provisioning=self.disable_provisioning,
                       disable_shared_folders=self.disable_shared_folders,
                       sep=sep))

    def config_ssh(self):
        """Configure ssh to work. If needed, create an insecure private key file for ssh/scp."""
        if not self.get_ip():
            sys.exit(colored.red(textwrap.fill(
                "This Mech machine is reporting that it is not yet ready for SSH. "
                "Make sure your machine is created and running and try again. "
                "Additionally, check the output of `mech ls` to verify "
                "that the machine is in the state that you expect.")))

        if not self.use_psk:
            key = os.path.abspath(os.path.join(
                utils.mech_dir(), "insecure_private_key"))
            if not os.path.exists(key):
                with open(key, 'w') as the_file:
                    the_file.write(INSECURE_PRIVATE_KEY)
                os.chmod(key, 0o400)
        else:
            key = '~/.ssh/id_rsa'

        self.config = {
            "Host": self.name,
            "User": self.user,
            "Port": "22",
            "UserKnownHostsFile": "/dev/null",
            "StrictHostKeyChecking": "no",
            "PasswordAuthentication": "no",
            "IdentityFile": key,
            "IdentitiesOnly": "yes",
            "LogLevel": "FATAL",
        }
        for key, value in self.config.items():
            key = re.sub(r'[ _]+', r' ', key)
            key = re.sub(r'(?<=[^_])([A-Z])', r' \1', key).lower()
            key = re.sub(r'^( *)(.*?)( *)$', r'\2', key)

            def callback(pat):
                return pat.group(1).upper()

            key = re.sub(r' (\w)', callback, key)
            if key[0].islower():
                key = key[0].upper() + key[1:]
            self.config[key] = value
        self.config.update({
            "HostName": self.ip,
        })
        return self.config



# Copyright (c) 2020 Mike Kinney

"""Test mech utils."""
import os
import re
import sys
import subprocess

from unittest.mock import patch, mock_open, MagicMock
from collections import OrderedDict
from pytest import raises

import mech.utils
import mech.mech_instance


@patch('os.getcwd')
def test_main_dir(mock_os_getcwd):
    """Test main_dir()."""
    mock_os_getcwd.return_value = '/tmp'
    main = mech.utils.main_dir()
    mock_os_getcwd.assert_called()
    assert main == '/tmp'


@patch('os.getcwd')
def test_mech_dir(mock_os_getcwd):
    """Test mech_dir()."""
    mock_os_getcwd.return_value = '/tmp'
    mechdir = mech.utils.mech_dir()
    mock_os_getcwd.assert_called()
    assert mechdir == '/tmp/.mech'


@patch('os.makedirs')
def test_makedirs(mock_os_makedirs):
    """Test makedirs()."""
    mock_os_makedirs.return_value = True
    mech.utils.makedirs('/tmp/1234')
    mock_os_makedirs.assert_called()


def test_confirm_yes():
    """Test confirm."""
    a_mock = MagicMock()
    a_mock.return_value = 'Y'
    with patch('mech.utils.raw_input', a_mock):
        assert mech.utils.confirm("Is this silly?")


def test_confirm_yes_by_default():
    """Test confirm."""
    a_mock = MagicMock()
    a_mock.return_value = ''
    with patch('mech.utils.raw_input', a_mock):
        assert mech.utils.confirm("Is this silly?")


def test_confirm_no_by_default():
    """Test confirm."""
    a_mock = MagicMock()
    a_mock.return_value = ''
    with patch('mech.utils.raw_input', a_mock):
        assert not mech.utils.confirm("Is this silly?", 'n')


def test_confirm_no():
    """Test confirm."""
    a_mock = MagicMock()
    a_mock.return_value = 'n'
    with patch('mech.utils.raw_input', a_mock):
        assert not mech.utils.confirm("Is this silly?")


def test_confirm_nonstandard_default():
    """Test confirm."""
    a_mock = MagicMock()
    a_mock.return_value = 'y'
    with patch('mech.utils.raw_input', a_mock):
        assert mech.utils.confirm("Is this silly?", 'q')


@patch('json.loads')
@patch('os.path.isfile')
@patch('os.getcwd')
def test_load_mechfile(mock_os_getcwd, mock_os_path_isfile, mock_json_loads):
    """Test mech_load_mechfile()."""
    mock_os_getcwd.return_value = '/tmp'
    expected = {}
    mock_json_loads.return_value = expected
    mock_os_path_isfile.return_value = True
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        assert mech.utils.load_mechfile() == expected
    a_mock.assert_called()
    mock_os_getcwd.assert_called()


@patch('json.loads')
@patch('os.path.isfile')
@patch('os.getcwd')
def test_load_mechfile_no_mechfile(mock_os_getcwd, mock_os_path_isfile, mock_json_loads):
    """Test mech_load_mechfile()."""
    mock_os_getcwd.return_value = '/tmp'
    expected = {}
    mock_json_loads.return_value = expected
    mock_os_path_isfile.return_value = False
    with raises(SystemExit):
        a_mock = mock_open()
        with patch('builtins.open', a_mock, create=True):
            mech.utils.load_mechfile()


@patch('json.loads')
@patch('os.path.isfile')
@patch('os.getcwd')
def test_load_mechfile_no_mechfile_should_not_exist(mock_os_getcwd, mock_os_path_isfile,
                                                    mock_json_loads):
    """Test mech_load_mechfile()."""
    mock_os_getcwd.return_value = '/tmp'
    expected = {}
    mock_json_loads.return_value = expected
    mock_os_path_isfile.return_value = False
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        got = mech.utils.load_mechfile(should_exist=False)
    assert got == expected
    mock_os_getcwd.assert_called()


@patch('os.path.isfile')
@patch('os.getcwd')
def test_load_mechfile_invalid_json(mock_os_getcwd, mock_os_path_isfile):
    """Test mech_load_mechfile()."""
    mock_os_getcwd.return_value = '/tmp'
    expected = {}
    # bad_json below is missing a comma between elements
    bad_json = '''{"foo": "bar" "foo2": 1}'''
    mock_os_path_isfile.return_value = True
    a_mock = mock_open(read_data=bad_json)
    with patch('builtins.open', a_mock, create=True):
        got = mech.utils.load_mechfile()
    assert got == expected
    a_mock.assert_called()
    mock_os_getcwd.assert_called()


def test_save_mechfile_empty_config():
    """Test save_mechfile with empty configuration."""
    filename = os.path.join(mech.utils.main_dir(), 'Mechfile')
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        assert mech.utils.save_mechfile({})
    a_mock.assert_called_once_with(filename, 'w+')
    a_mock.return_value.write.assert_called_once_with('{}')


def test_save_mechfile_one(helpers):
    """Test save_mechfile with one entry."""
    first_dict = {
        'first': {
            'name':
            'first',
            'box':
            'bento/ubuntu-18.04',
            'box_version':
            '201912.04.0',
            'url':
            'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/'
            'versions/201912.04.0/providers/vmware_desktop.box'
        }
    }
    first_json = '''{
  "first": {
    "box": "bento/ubuntu-18.04",
    "box_version": "201912.04.0",
    "name": "first",
    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/versions/201912.04.0/providers/vmware_desktop.box"
  }
}'''  # noqa: 501
    filename = os.path.join(mech.utils.main_dir(), 'Mechfile')
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        assert mech.utils.save_mechfile(first_dict)
    a_mock.assert_called_once_with(filename, 'w+')
    assert first_json == helpers.get_mock_data_written(a_mock)


def test_save_mechfile_two(helpers):
    """Test save_mechfile with two entries."""
    two_dict = {
        'first': {
            'name':
            'first',
            'box':
            'bento/ubuntu-18.04',
            'box_version':
            '201912.04.0',
            'url':
            'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/'
            'versions/201912.04.0/providers/vmware_desktop.box'
        },
        'second': {
            'name':
            'second',
            'box':
            'bento/ubuntu-18.04',
            'box_version':
            '201912.04.0',
            'url':
            'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/'
            'versions/201912.04.0/providers/vmware_desktop.box'
        }
    }
    two_json = '''{
  "first": {
    "box": "bento/ubuntu-18.04",
    "box_version": "201912.04.0",
    "name": "first",
    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/versions/201912.04.0/providers/vmware_desktop.box"
  },
  "second": {
    "box": "bento/ubuntu-18.04",
    "box_version": "201912.04.0",
    "name": "second",
    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/versions/201912.04.0/providers/vmware_desktop.box"
  }
}'''  # noqa: 501
    filename = os.path.join(mech.utils.main_dir(), 'Mechfile')
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        assert mech.utils.save_mechfile(two_dict)
    a_mock.assert_called_once_with(filename, 'w+')
    assert two_json == helpers.get_mock_data_written(a_mock)


@patch('os.getlogin')
@patch('mech.utils.save_mechfile_entry')
@patch('mech.utils.build_mechfile_entry')
def test_init_mechfile_with_add_me_option(mock_build_mechfile_entry, mock_save_mechfile_entry,
                                          mock_os_getlogin):
    """Test init_mechfile."""
    mock_build_mechfile_entry.return_value = {}
    mock_save_mechfile_entry.return_value = True
    mock_os_getlogin.return_value = 'bob'
    assert mech.utils.init_mechfile(add_me=True)
    mock_build_mechfile_entry.assert_called()
    mock_save_mechfile_entry.assert_called()


@patch('os.getlogin')
@patch('mech.utils.save_mechfile_entry')
@patch('mech.utils.build_mechfile_entry')
def test_add_to_mechfile_with_add_me_option(mock_build_mechfile_entry, mock_save_mechfile_entry,
                                            mock_os_getlogin):
    """Test add_to_mechfile."""
    mock_build_mechfile_entry.return_value = {}
    mock_save_mechfile_entry.return_value = True
    mock_os_getlogin.return_value = 'bob'
    assert mech.utils.add_to_mechfile(add_me=True)
    mock_build_mechfile_entry.assert_called()
    mock_save_mechfile_entry.assert_called()


@patch('mech.vmrun.VMrun.run_script_in_guest', return_value='')
@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth(mock_locate, mock_installed_tools,
                  mock_run_script_in_guest, capfd,
                  mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    mock_installed_tools.return_value = "running"
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    print('inst:{}'.format(inst))
    pub_key = 'some fake pub key'
    a_mock = mock_open(read_data=pub_key)
    with patch('builtins.open', a_mock, create=True):
        mech.utils.add_auth(inst)
        a_mock.assert_called()
        out, _ = capfd.readouterr()
        mock_locate.assert_called()
        mock_installed_tools.assert_called()
        assert re.search(r'Adding auth', out, re.MULTILINE)
        assert re.search(r'Added auth', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.run_script_in_guest', return_value=None)
@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth_script_fails(mock_locate, mock_installed_tools,
                               mock_run_script_in_guest, capfd,
                               mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    mock_installed_tools.return_value = "running"
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    print('inst:{}'.format(inst))
    pub_key = 'some fake pub key'
    a_mock = mock_open(read_data=pub_key)
    with patch('builtins.open', a_mock, create=True):
        mech.utils.add_auth(inst)
        a_mock.assert_called()
        out, _ = capfd.readouterr()
        mock_locate.assert_called()
        mock_installed_tools.assert_called()
        assert re.search(r'Adding auth', out, re.MULTILINE)
        assert re.search(r'Did not add auth', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.run_script_in_guest', return_value=None)
@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth_could_not_read_pub_key(mock_locate, mock_installed_tools,
                                         mock_run_script_in_guest, capfd,
                                         mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    mock_installed_tools.return_value = "running"
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    print('inst:{}'.format(inst))
    a_mock = mock_open(read_data=None)
    with patch('builtins.open', a_mock, create=True):
        mech.utils.add_auth(inst)
        a_mock.assert_called()
        out, _ = capfd.readouterr()
        mock_locate.assert_called()
        mock_installed_tools.assert_called()
        assert re.search(r'Adding auth', out, re.MULTILINE)
        assert re.search(r'Could not read contents', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.run_script_in_guest', return_value=None)
@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth_no_username_in_auth(mock_locate, mock_installed_tools,
                                      mock_run_script_in_guest, capfd,
                                      mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    mock_installed_tools.return_value = "running"
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    # remove username
    del inst.auth['username']
    print('inst:{}'.format(inst))
    mech.utils.add_auth(inst)
    out, _ = capfd.readouterr()
    mock_locate.assert_called()
    mock_installed_tools.assert_called()
    assert re.search(r'Adding auth', out, re.MULTILINE)
    assert re.search(r'Warning: Need a username', out, re.MULTILINE)


@patch('mech.vmrun.VMrun.run_script_in_guest', return_value=None)
@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth_no_auth(mock_locate, mock_installed_tools,
                          mock_run_script_in_guest, capfd,
                          mechfile_one_entry):
    """Test add_auth."""
    mock_installed_tools.return_value = "running"
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    print('inst:{}'.format(inst))
    mech.utils.add_auth(inst)
    out, _ = capfd.readouterr()
    mock_locate.assert_called()
    mock_installed_tools.assert_called()
    assert re.search(r'Adding auth', out, re.MULTILINE)
    assert re.search(r'No auth to add', out, re.MULTILINE)


def test_add_auth_no_instance():
    """Test add_auth when no instance."""
    with raises(SystemExit, match=r"Need to provide an instance"):
        mech.utils.add_auth(None)


def test_add_auth_no_vmx(mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = ''
    inst.password = 'vagrant'
    inst.vmx = None
    with raises(SystemExit, match=r"Need to provide vmx"):
        mech.utils.add_auth(inst)


def test_add_auth_blank_user(mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = ''
    inst.password = 'vagrant'
    inst.vmx = '/tmp/first/some.vmx'
    with raises(SystemExit, match=r"Need to provide user"):
        mech.utils.add_auth(inst)


def test_add_auth_password_is_none(mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = None
    inst.vmx = '/tmp/first/some.vmx'
    with raises(SystemExit, match=r"Need to provide password"):
        mech.utils.add_auth(inst)


def test_add_auth_blank_passwod(mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = ''
    inst.vmx = '/tmp/first/some.vmx'
    with raises(SystemExit, match=r"Need to provide password"):
        mech.utils.add_auth(inst)


def test_add_auth_user_is_none(mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = None
    inst.password = 'vagrant'
    inst.vmx = '/tmp/first/some.vmx'
    with raises(SystemExit, match=r"Need to provide user"):
        mech.utils.add_auth(inst)


@patch('mech.vmrun.VMrun.installed_tools')
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_add_auth_cannot_add_auth(mock_locate, mock_installed_tools,
                                  capfd,
                                  mechfile_one_entry_with_auth_and_mech_use):
    """Test add_auth."""
    mock_installed_tools.return_value = False
    inst = mech.mech_instance.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    inst.user = 'vagrant'
    inst.password = 'vagrant'
    with raises(SystemExit, match=r"Cannot add auth"):
        mech.utils.add_auth(inst)


def test_tar_cmd():
    """Test tar cmd.
    """
    a_mock = MagicMock()
    another_mock = MagicMock()
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    yet_another_mock.return_value = 'blah blah blah --force-local boo --fast-read blah', None
    another_mock.communicate = yet_another_mock
    another_mock.communicate.returncode = 0
    another_mock.returncode = 0
    a_mock.return_value = another_mock
    a_mock.returncode = 0
    if sys.platform.startswith('darwin'):
        expected = ["tar", "--force-local", "--fast-read"]
    else:
        expected = ["tar", "--force-local"]
    with patch('subprocess.Popen', a_mock):
        got = mech.utils.tar_cmd(wildcards=True, fast_read=True, force_local=True)
        assert expected == got


def test_tar_cmd_when_tar_not_found():
    """Test tar cmd."""
    a_mock = MagicMock()
    a_mock.return_value = None
    a_mock.returncode = None
    a_mock.side_effect = OSError()
    with patch('subprocess.Popen', a_mock):
        tar = mech.utils.tar_cmd()
        assert tar is None


def test_config_ssh_string_empty():
    """Test config_ssh_string with empty configuration."""
    ssh_string = mech.utils.config_ssh_string({})
    assert ssh_string == "Host \n"


def test_config_ssh_string_simple():
    """Test config_ssh_string with a simple configuration."""
    config = {
        "Host": "first",
        "User": "foo",
        "Port": "22",
        "UserKnownHostsFile": "/dev/null",
        "StrictHostKeyChecking": "no",
        "PasswordAuthentication": "no",
        "IdentityFile": 'blah',
        "IdentitiesOnly": "yes",
        "LogLevel": "FATAL",
    }
    ssh_string = mech.utils.config_ssh_string(config)
    assert ssh_string == 'Host first\n  User foo\n  Port 22\n  UserKnownHostsFile /dev/null\n  StrictHostKeyChecking no\n  PasswordAuthentication no\n  IdentityFile blah\n  IdentitiesOnly yes\n  LogLevel FATAL\n'  # noqa: E501  pylint: disable=line-too-long


@patch('mech.utils.load_mechfile', return_value={})
@patch('mech.utils.save_mechfile', return_value=True)
def test_save_mechfile_entry_with_empty_mechfile(load_mock, save_mock):
    """Test save_mechfile_entry with no entries in the mechfile."""
    entry = {'first': {'name': 'first'}}
    assert mech.utils.save_mechfile_entry(entry, 'first', True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()


@patch('mech.utils.load_mechfile', return_value={})
@patch('mech.utils.save_mechfile', return_value=True)
def test_save_mechfile_entry_with_blank_name(load_mock, save_mock):
    """Test save_mechfile_entry with a blank name."""
    entry = {'first': {'name': 'first'}}
    assert mech.utils.save_mechfile_entry(entry, '', True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()


@patch('mech.utils.load_mechfile', return_value={})
@patch('mech.utils.save_mechfile', return_value=True)
def test_save_mechfile_entry_with_name_as_none(load_mock, save_mock):
    """Test save_mechfile_entry with name as None."""
    entry = {'first': {'name': 'first'}}
    assert mech.utils.save_mechfile_entry(entry, None, True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()


@patch('mech.utils.load_mechfile', return_value={})
@patch('mech.utils.save_mechfile', return_value=True)
def test_save_mechfile_entry_twice(load_mock, save_mock):
    """Test save_mechfile_entry multiple times."""
    entry = {'first': {'name': 'first'}}
    assert mech.utils.save_mechfile_entry(entry, 'first', True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()
    assert mech.utils.save_mechfile_entry(entry, 'first', True)


@patch('mech.utils.load_mechfile', return_value={})
@patch('mech.utils.save_mechfile', return_value=True)
def test_remove_mechfile_entry_with_empty_mechfile(load_mock, save_mock):
    """Test remove_mechfile_entry with no entries in the mechfile."""
    assert mech.utils.remove_mechfile_entry('first', True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()


@patch('mech.utils.load_mechfile', return_value={'first': {'name': 'first'}})
@patch('mech.utils.save_mechfile', return_value=True)
def test_remove_mechfile_entry(load_mock, save_mock):
    """Test remove_mechfile_entry."""
    assert mech.utils.remove_mechfile_entry('first', True)
    load_mock.assert_called_once()
    save_mock.assert_called_once()


def test_parse_vmx():
    """Test parse_vmx."""
    partial_vmx = '''.encoding = "UTF-8"
bios.bootorder  = "hdd,cdrom"
checkpoint.vmstate     = ""

cleanshutdown = "FALSE"
config.version = "8"'''
    expected_vmx = OrderedDict([
        ('.encoding', '"UTF-8"'),
        ('bios.bootorder', '"hdd,cdrom"'),
        ('checkpoint.vmstate', '""'),
        ('cleanshutdown', '"FALSE"'),
        ('config.version', '"8"')
    ])
    a_mock = mock_open(read_data=partial_vmx)
    with patch('builtins.open', a_mock):
        assert mech.utils.parse_vmx(partial_vmx) == expected_vmx
    a_mock.assert_called()


@patch('mech.utils.parse_vmx')
def test_update_vmx_empty(mock_parse_vmx, helpers, capfd):
    """Test update_vmx."""
    expected_vmx = """ethernet0.addresstype = generated
ethernet0.bsdname = en0
ethernet0.connectiontype = nat
ethernet0.displayname = Ethernet
ethernet0.linkstatepropagation.enable = FALSE
ethernet0.pcislotnumber = 32
ethernet0.present = TRUE
ethernet0.virtualdev = e1000
ethernet0.wakeonpcktrcv = FALSE
"""
    mock_parse_vmx.return_value = {}
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        mech.utils.update_vmx('/tmp/first/one.vmx')
        a_mock.assert_called()
        got = helpers.get_mock_data_written(a_mock)
        assert expected_vmx == got
        out, _ = capfd.readouterr()
        assert re.search(r'Added network interface to vmx file', out, re.MULTILINE)


@patch('mech.utils.parse_vmx')
def test_update_vmx_with_a_network_entry(mock_parse_vmx, capfd):
    """Test update_vmx."""
    mock_parse_vmx.return_value = {'ethernet0.present': 'true'}
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        mech.utils.update_vmx('/tmp/first/one.vmx')
        assert not a_mock.called, 'should not have written anything to the vmx file'
        out, _ = capfd.readouterr()
        assert out == ''


@patch('mech.utils.parse_vmx')
def test_update_vmx_with_cpu_and_memory(mock_parse_vmx, helpers, capfd):
    """Test update_vmx."""
    mock_parse_vmx.return_value = {'ethernet0.present': 'true'}
    expected_vmx = '''ethernet0.present = true
numvcpus = "3"
memsize = "1025"
'''
    a_mock = mock_open()
    with patch('builtins.open', a_mock, create=True):
        mech.utils.update_vmx('/tmp/first/one.vmx', numvcpus=3, memsize=1025)
        a_mock.assert_called()
        got = helpers.get_mock_data_written(a_mock)
        assert expected_vmx == got
        out, _ = capfd.readouterr()
        assert out == ''


def test_build_mechfile_entry_no_location():
    """Test if None is used for location."""
    assert mech.utils.build_mechfile_entry(location=None) == {}


def test_build_mechfile_entry_https_location():
    """Test if location starts with 'https://'."""
    assert mech.utils.build_mechfile_entry(location='https://foo') == {
        'box': None,
        'box_version': None,
        'name': None,
        'provider': None,
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url': 'https://foo'
    }


def test_build_mechfile_entry_http_location():
    """Test if location starts with 'http://'."""
    assert mech.utils.build_mechfile_entry(location='http://foo') == {
        'box': None,
        'box_version': None,
        'name': None,
        'provider': None,
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url':
        'http://foo'
    }


def test_build_mechfile_entry_ftp_location():
    """Test if location starts with 'ftp://'."""
    assert mech.utils.build_mechfile_entry(location='ftp://foo') == {
        'box': None,
        'box_version': None,
        'name': None,
        'provider': None,
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url': 'ftp://foo'
    }


def test_build_mechfile_entry_ftp_location_with_other_values():
    """Test if mechfile_entry is filled out."""
    expected = {
        'box': 'bbb',
        'box_version': 'ccc',
        'name': 'aaa',
        'provider': 'vmware',
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url': 'ftp://foo'
    }
    assert mech.utils.build_mechfile_entry(location='ftp://foo', name='aaa',
                                           box='bbb', box_version='ccc',
                                           provider='vmware') == expected


def test_build_mechfile_entry_file_location_json(catalog):
    """Test if location starts with 'file:' and contains valid json."""

    # Note: Download/format json like this:
    # curl --header 'Accept:application/json' \
    #    'https://app.vagrantup.com/bento/boxes/ubuntu-18.04' | python3 -m json.tool
    expected = {
        'box': 'bento/ubuntu-18.04',
        'box_version': 'aaa',
        'name': 'first',
        'provider': 'vmware',
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url':
        'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/aaa/providers/vmware_desktop.box'
    }
    a_mock = mock_open(read_data=catalog)
    with patch('builtins.open', a_mock):
        actual = mech.utils.build_mechfile_entry(location='file:/tmp/one.json')
        assert expected == actual
    a_mock.assert_called()


def test_build_mechfile_entry_file_location_but_file_not_found():
    """Test if location starts with 'file:' and file does not exist."""
    with patch('builtins.open', mock_open()) as mock_file:
        mock_file.side_effect = SystemExit()
        with raises(SystemExit):
            mech.utils.build_mechfile_entry(location='file:/tmp/one.box')


@patch('requests.get')
def test_build_mechfile_entry_file_location_external_good(mock_requests_get,
                                                          catalog_as_json):
    """Test if location talks to Hashicorp."""
    expected = {
        'box': 'bento/ubuntu-18.04',
        'box_version': 'aaa',
        'name': None,
        'provider': 'vmware',
        'shared_folders': [{'host_path': '.', 'share_name': 'mech'}],
        'url':
        'https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/aaa/providers/vmware_desktop.box'
    }
    mock_requests_get.return_value.status_code = 200
    mock_requests_get.return_value.json.return_value = catalog_as_json
    actual = mech.utils.build_mechfile_entry(location='bento/ubuntu-18.04')
    mock_requests_get.assert_called()
    assert expected == actual


def test_build_mechfile_entry_file_location_external_bad_location():
    """Test if we do not have a valid location. (must be in form of 'hashiaccount/boxname')."""
    with raises(SystemExit, match=r"Provided box name is not valid"):
        mech.utils.build_mechfile_entry(location='bento')


def test_deL_user_no_instance():
    """Test del_user."""
    with raises(SystemExit, match=r"Need to provide an instance"):
        mech.utils.del_user(None, 'bart')


def test_deL_user_vm_not_created():
    """Test del_user."""
    mock_inst = MagicMock()
    mock_inst.vmx = None
    with raises(SystemExit, match=r"VM must be created"):
        mech.utils.del_user(mock_inst, 'bart')


def test_deL_user_no_username_to_use_for_auth():
    """Test del_user."""
    mock_inst = MagicMock()
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.user = None
    with raises(SystemExit, match=r"A user is required"):
        mech.utils.del_user(mock_inst, 'bart')


def test_deL_user_no_user_provided():
    """Test del_user."""
    mock_inst = MagicMock()
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.user = 'bob'
    with raises(SystemExit, match=r"A username to delete is required"):
        mech.utils.del_user(mock_inst, None)


@patch('mech.utils.ssh', return_value='')
def test_deL_user(mock_ssh):
    """Test del_user."""
    mock_inst = MagicMock()
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.user = 'bob'
    mech.utils.del_user(mock_inst, 'homer')


def test_provision_no_instance():
    """Test provisioning."""
    with raises(SystemExit, match=r"Need to provide an instance to provision"):
        mech.utils.provision(instance=None, show=None)


def test_provision_no_vmx():
    """Test provisioning."""
    mock_inst = MagicMock()
    mock_inst.vmx = None
    mock_inst.provider = 'vmware'
    with raises(SystemExit, match=r"Need to provide vmx.*"):
        mech.utils.provision(instance=mock_inst, show=None)


def test_provision_no_vbox():
    """Test provisioning."""
    mock_inst = MagicMock()
    mock_inst.vbox = None
    mock_inst.provider = 'virtualbox'
    with raises(SystemExit, match=r"Need to provide vbox.*"):
        mech.utils.provision(instance=mock_inst, show=None)


@patch('mech.utils.scp', return_value='')
def test_provision_file(mock_scp):
    """Test provision_file"""
    mock_inst = MagicMock()
    mech.utils.provision_file(mock_inst, 'somefile', 'somedest')
    mock_scp.assert_called()


@patch('mech.utils.ssh')
def test_create_tempfile_in_guest(mock_ssh):
    """Test create_tempfile_in_guest"""
    expected = '/tmp/foo123'
    mock_ssh.return_value = None, expected, None
    mock_inst = MagicMock()
    got = mech.utils.create_tempfile_in_guest(mock_inst)
    mock_ssh.assert_called()
    assert got == expected


@patch('mech.vmrun.VMrun.installed_tools')
def test_provision_file_show(mock_installed_tools, capfd):
    """Test provisioning."""
    mock_installed_tools.return_value = "running"
    config = [
        {
            "type": "file",
            "source": "file1.txt",
            "destination": "/tmp/file1.txt",
        },
    ]
    mock_inst = MagicMock()
    mock_inst.name = 'first'
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.provision = config
    mech.utils.provision(instance=mock_inst, show=True)
    out, _ = capfd.readouterr()
    assert re.search(r'instance:', out, re.MULTILINE)


def test_provision_which_has_shell_show_only(capfd):
    """Test provisioning."""
    config = [
        {
            "type": "shell",
            "path": "file1.sh",
            "args": [
                "a=1",
                "b=true",
            ],
        },
    ]
    mock_inst = MagicMock()
    mock_inst.name = 'first'
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.provision = config
    mech.utils.provision(instance=mock_inst, show=True)
    out, _ = capfd.readouterr()
    assert re.search(r' instance:', out, re.MULTILINE)


@patch('mech.utils.provision_shell', return_value=None)
def test_provision_that_has_shell_with_issue(mock_provision_shell,
                                             capfd):
    """Test provisioning."""
    config = [
        {
            "type": "shell",
            "path": "file1.sh",
            "args": [
                "a=1",
                "b=true",
            ],
        },
    ]
    mock_inst = MagicMock()
    mock_inst.name = 'first'
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.provision = config
    mech.utils.provision(instance=mock_inst, show=None)
    out, _ = capfd.readouterr()
    mock_provision_shell.assert_called()
    assert re.search(r'Not Provisioned', out, re.MULTILINE)


def test_provision_with_unknown_type(capfd):
    """Test provisioning."""
    config = [
        {
            "type": "foo",
        },
    ]
    mock_inst = MagicMock()
    mock_inst.name = 'first'
    mock_inst.vmx = '/tmp/first/some.vmx'
    mock_inst.provision = config
    mech.utils.provision(instance=mock_inst, show=None)
    out, _ = capfd.readouterr()
    assert re.search(r'Not Provisioned', out, re.MULTILINE)


@patch('mech.utils.ssh', return_value=[True, True, True])
@patch('mech.utils.scp', return_value=True)
@patch('os.path.isfile', return_value=True)
@patch('mech.utils.create_tempfile_in_guest', return_value='/tmp/foo')
def test_provision_shell(mock_create_tempfile, mock_isfile,
                         mock_scp, mock_ssh, capfd,
                         mechfile_one_entry_with_auth_and_mech_use):
    """Test provision_shell."""
    inst = mech.mech_instance.MechInstance('first',
                                           mechfile_one_entry_with_auth_and_mech_use)
    some_vmx = '/tmp/first/some.vmx'
    inst.vmx = some_vmx
    inst.created = True
    mech.utils.provision_shell(inst, inline=False, script_path='file1.sh', args=None)
    out, _ = capfd.readouterr()
    mock_create_tempfile.assert_called()
    mock_isfile.assert_called()
    mock_scp.assert_called()
    mock_ssh.assert_called()
    assert re.search(r'Configuring script', out, re.MULTILINE)
    assert re.search(r'Configuring environment', out, re.MULTILINE)
    assert re.search(r'Executing program', out, re.MULTILINE)


@patch('mech.utils.ssh', return_value=None)
@patch('mech.utils.scp', return_value=True)
@patch('os.path.isfile', return_value=True)
@patch('mech.utils.create_tempfile_in_guest', return_value='/tmp/foo')
def test_provision_shell_make_executable_fails(mock_create_tempfile, mock_isfile,
                                               mock_scp, mock_ssh, capfd,
                                               mechfile_one_entry_with_auth_and_mech_use):
    """Test provision_shell."""
    inst = mech.mech_instance.MechInstance('first',
                                           mechfile_one_entry_with_auth_and_mech_use)
    some_vmx = '/tmp/first/some.vmx'
    inst.vmx = some_vmx
    inst.created = True
    mech.utils.provision_shell(inst, inline=False, script_path='file1.sh', args=None)
    out, _ = capfd.readouterr()
    mock_create_tempfile.assert_called()
    mock_isfile.assert_called()
    mock_scp.assert_called()
    mock_ssh.assert_called()
    assert re.search(r'Configuring script', out, re.MULTILINE)
    assert re.search(r'Configuring environment', out, re.MULTILINE)
    assert re.search(r'Warning: Could not configure script', out, re.MULTILINE)


@patch('requests.get')
@patch('mech.utils.ssh', return_value=[True, True, True])
@patch('mech.utils.scp', return_value=True)
@patch('os.path.isfile', return_value=False)
@patch('mech.utils.create_tempfile_in_guest', return_value='/tmp/foo')
def test_provision_shell_from_http(mock_create_tempfile, mock_isfile,
                                   mock_scp, mock_ssh, mock_requests_get,
                                   capfd,
                                   mechfile_one_entry_with_auth_and_mech_use):
    """Test provision_shell."""
    inst = mech.mech_instance.MechInstance('first',
                                           mechfile_one_entry_with_auth_and_mech_use)
    some_vmx = '/tmp/first/some.vmx'
    inst.vmx = some_vmx
    inst.created = True
    mock_requests_get.return_value.status_code = 200
    mock_requests_get.return_value.raise_for_status.return_value = None
    mock_requests_get.return_value.read.return_value = 'echo hello'
    mech.utils.provision_shell(inst, inline=False, script_path='http://example.com/file1.sh',
                               args=None)
    out, _ = capfd.readouterr()
    mock_create_tempfile.assert_called()
    mock_isfile.assert_called()
    mock_scp.assert_called()
    mock_ssh.assert_called()
    mock_requests_get.assert_called()
    assert re.search(r'Configuring script', out, re.MULTILINE)
    assert re.search(r'Configuring environment', out, re.MULTILINE)
    assert re.search(r'Executing program', out, re.MULTILINE)


@patch('mech.utils.ssh', return_value=[True, True, True])
@patch('mech.utils.scp', return_value=None)
@patch('os.path.isfile', return_value=True)
@patch('mech.utils.create_tempfile_in_guest', return_value='/tmp/foo')
def test_provision_shell_cannot_copy_file(mock_create_tempfile, mock_isfile,
                                          mock_scp, mock_ssh,
                                          capfd,
                                          mechfile_one_entry_with_auth_and_mech_use):
    """Test provision_shell."""
    inst = mech.mech_instance.MechInstance('first',
                                           mechfile_one_entry_with_auth_and_mech_use)
    some_vmx = '/tmp/first/some.vmx'
    inst.vmx = some_vmx
    inst.created = True
    mech.utils.provision_shell(inst, inline=False, script_path='file1.sh', args=None)
    out, _ = capfd.readouterr()
    mock_create_tempfile.assert_called()
    mock_isfile.assert_called()
    mock_ssh.assert_called()
    mock_scp.assert_called()
    assert re.search(r'Configuring script', out, re.MULTILINE)
    assert re.search(r'Warning: Could not copy file to guest', out, re.MULTILINE)


@patch('os.environ')
def test_get_fallback_executable_no_path_in_environ(mock_os_environ):
    """Weird case where PATH is is not in the environment."""
    mock_os_environ.return_value = ''
    assert mech.utils.get_fallback_executable() is None


def test_run_pyinfra_script_no_host():
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host=None, username=None) is None


def test_run_pyinfra_script_no_username():
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username=None) is None


def test_run_pyinfra_script_script_is_none():
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username='vagrant') is None


def test_run_pyinfra_script_script_no_script():
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username='vagrant', script_path='') is None


@patch('os.path.exists', return_value=False)
def test_run_pyinfra_script_script_not_found(mock_path_exists):
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username='vagrant',
                                         script_path='/tmp/file1.py') is None
    mock_path_exists.assert_called()


@patch('os.path.exists', return_value=True)
def test_run_pyinfra_script_script_does_not_end_in_py(mock_path_exists):
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username='vagrant',
                                         script_path='/tmp/file1') is None
    mock_path_exists.assert_called()


@patch('mech.utils.pyinfra_installed', return_value=False)
@patch('os.path.exists', return_value=True)
def test_run_pyinfra_script_pyinfra_not_installed(mock_path_exists, mock_pyinfra_installed):
    """Test pyinfra_script"""
    assert mech.utils.run_pyinfra_script(host='foo', username='vagrant',
                                         script_path='/tmp/file1.py') is None
    mock_path_exists.assert_called()
    mock_pyinfra_installed.assert_called()


@patch('mech.utils.pyinfra_installed', return_value=True)
@patch('os.path.exists', return_value=True)
def test_run_pyinfra_script_pyinfra_success(mock_path_exists, mock_pyinfra_installed):
    """Test pyinfra_script"""
    mock_subprocess = MagicMock()
    mock_subprocess.return_value = subprocess.CompletedProcess(args='',
                                                               returncode=0,
                                                               stdout=b'',
                                                               stderr=b'')
    with patch('subprocess.run', mock_subprocess):
        return_code, stdout, stderr = mech.utils.run_pyinfra_script(host='foo', username='vagrant',
                                                                    script_path='/tmp/file1.py')
        assert return_code == 0
        assert stdout == ''
        assert stderr == ''
        mock_path_exists.assert_called()
        mock_pyinfra_installed.assert_called()
        mock_subprocess.assert_called()


@patch('mech.utils.pyinfra_installed', return_value=True)
@patch('os.path.exists', return_value=True)
def test_run_pyinfra_script_pyinfra_failed(mock_path_exists, mock_pyinfra_installed):
    """Test pyinfra_script"""
    mock_subprocess = MagicMock()
    mock_subprocess.return_value = subprocess.CompletedProcess(args='',
                                                               returncode=1,
                                                               stdout=b'some output',
                                                               stderr=b'some error')
    with patch('subprocess.run', mock_subprocess):
        return_code, stdout, stderr = mech.utils.run_pyinfra_script(host='foo', username='vagrant',
                                                                    script_path='/tmp/file1.py')
        assert return_code == 1
        assert stdout == 'some output'
        assert stderr == 'some error'
        mock_path_exists.assert_called()
        mock_pyinfra_installed.assert_called()
        mock_subprocess.assert_called()


def test_pyinfra_installed():
    """Test pyinfra_installed"""
    mock_subprocess = MagicMock()
    mock_subprocess.return_value = subprocess.CompletedProcess(args='', returncode=0)
    with patch('subprocess.run', mock_subprocess):
        assert mech.utils.pyinfra_installed()


def test_pyinfra_not_installed():
    """Test pyinfra_installed"""
    mock_subprocess = MagicMock()
    mock_subprocess.return_value = subprocess.CompletedProcess(args='', returncode=127)
    with patch('subprocess.run', mock_subprocess):
        assert not mech.utils.pyinfra_installed()


@patch('os.path.exists')
def test_get_fallback_executable(mock_os_path_exists):
    """Find vmrun in PATH."""
    mock_os_path_exists.return_value = True
    with patch.dict('os.environ', {'PATH': '/tmp:/tmp2'}):
        got = mech.utils.get_fallback_executable()
    expected = '/tmp/vmrun'
    assert got == expected
    mock_os_path_exists.assert_called()


@patch('os.path.exists')
def test_darwin_executable_when_installed(mock_os_path_exists):
    """Find vmrun in PATH."""
    expected = '/Applications/VMware Fusion.app/Contents/Library/vmrun'
    mock_os_path_exists.return_value = True
    got = mech.utils.get_darwin_executable()
    assert expected == got
    mock_os_path_exists.assert_called()


@patch('os.path.exists')
def test_darwin_executable_when_not_installed(mock_os_path_exists):
    """Find vmrun in PATH."""
    # deal with a different file returns a different mocked value
    def side_effect(filename):
        if filename == '/Applications/VMware Fusion.app/Contents/Library/vmrun':
            return False
        else:
            return True
    mock_os_path_exists.side_effect = side_effect
    expected = '/tmp/vmrun'
    with patch.dict('os.environ', {'PATH': '/tmp:/tmp2'}):
        got = mech.utils.get_darwin_executable()
    assert expected == got


@patch('os.path.exists')
def test_get_fallback_executable_when_exe(mock_os_path_exists):
    """Find vmrun.exe in PATH."""
    # deal with a different file returns a different mocked value
    def side_effect(filename):
        if filename == '/tmp/vmrun.exe':
            return True
        else:
            return False
    mock_os_path_exists.side_effect = side_effect
    expected = '/tmp/vmrun.exe'
    with patch.dict('os.environ', {'PATH': '/tmp:/tmp2'}):
        got = mech.utils.get_fallback_executable()
    assert expected == got


def test_catalog_to_mechfile_when_empty_catalog():
    """Test catalog_to_mechfile."""
    catalog = {}
    with raises(SystemExit):
        mech.utils.catalog_to_mechfile(catalog)


@patch('mech.utils.locate')
def test_init_box_cannot_find_valid_box(mock_locate):
    """Test init_box."""
    mock_locate.return_value = None
    with raises(SystemExit):
        mech.utils.init_box(name='first')


@patch('mech.utils.makedirs', return_value=True)
@patch('mech.utils.add_box')
@patch('mech.utils.locate')
def test_init_box_cannot_extract_box(mock_locate, mock_add_box, mock_makedirs):
    """Test init_box."""
    a_mock = MagicMock()
    another_mock = MagicMock()
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    yet_another_mock.return_value = 'someoutput', None
    another_mock.communicate = yet_another_mock
    another_mock.communicate.returncode = 0
    another_mock.returncode = 0
    a_mock.return_value = another_mock
    a_mock.returncode = 0
    mock_locate.return_value = None
    mock_add_box.return_value = 'bento', '1.23', 'ubuntu'
    with raises(SystemExit, match=r"Cannot extract box"):
        with patch('subprocess.Popen', a_mock):
            mech.utils.init_box(name='first', box='bento/ubuntu', box_version='1.23')


@patch('tarfile.open')
@patch('mech.utils.makedirs', return_value=True)
@patch('mech.utils.add_box')
@patch('mech.utils.locate')
def test_init_box_when_no_vmx_after_extraction(mock_locate, mock_add_box,
                                               mock_makedirs, mock_tarfile_open):
    """Test init_box."""
    mock_subprocess_popen = MagicMock()
    mock_tarfile_open = MagicMock()
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    yet_another_mock.return_value = 'someoutput', None
    mock_subprocess_popen.returncode = 0
    mock_locate.return_value = False
    mock_add_box.return_value = 'bento', '1.23', 'ubuntu'
    with raises(SystemExit, match=r"Cannot locate a VMX"):
        with patch('subprocess.Popen', mock_subprocess_popen):
            mech.utils.init_box(name='first', box='bento/ubuntu', box_version='1.23')
            mock_locate.assert_called()
            mock_add_box.assert_called()
            mock_makedirs.assert_called()
            mock_tarfile_open.assert_called()


@patch('mech.utils.update_vmx', return_value='/tmp/first/some.vmx')
@patch('tarfile.open')
@patch('mech.utils.makedirs', return_value=True)
@patch('mech.utils.add_box')
@patch('mech.utils.locate')
def test_init_box_success(mock_locate, mock_add_box, mock_makedirs,
                          mock_tarfile_open, mock_update_vmx):
    """Test init_box."""
    mock_subprocess_popen = MagicMock()
    mock_tarfile_open = MagicMock()
    mock_tarfile_open.returncode = 0
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    yet_another_mock.return_value = 'someoutput', None
    mock_subprocess_popen.returncode = 0
    # pytest tip: If you need to mock the same function multiple times,
    # put it into a list like the next line. Then mech.utils.locate() is
    # called the first time, it returns False. Subsequent two calls
    # return True. Mind blown.
    mock_locate.side_effect = [False, True, True]
    mock_add_box.return_value = 'bento', '1.23', 'ubuntu'
    with patch('subprocess.Popen', mock_subprocess_popen):
        mech.utils.init_box(name='first', box='bento/ubuntu', box_version='1.23')
        mock_locate.assert_called()
        mock_add_box.assert_called()
        mock_makedirs.assert_called()
        mock_update_vmx.assert_called()


@patch('mech.utils.mech_dir')
@patch('mech.utils.copyfile')
@patch('mech.utils.makedirs')
@patch('mech.utils.tar_cmd')
def test_add_box_file(mock_tar_cmd, mock_makedirs, mock_copyfile, mock_mech_dir):
    """Test add_box_file."""
    mock_tarfile_open = MagicMock()
    mock_tarfile_open.returncode = 0
    mock_tarfile_open.return_value.getnames.return_value = ['aaa', 'some.vmx']
    mock_tar_cmd.return_value = None
    mock_mech_dir.return_value = '/tmp'
    with patch('tarfile.open', mock_tarfile_open, create=True):
        box, box_version = mech.utils.add_box_file(box='bento/ubuntu',
                                                   box_version='1.23',
                                                   filename='/tmp/foo.box')
        mock_tarfile_open.assert_called()
        mock_tar_cmd.assert_called()
        mock_makedirs.assert_called()
        mock_copyfile.assert_called()
        assert box == '/tmp/boxes/bento/ubuntu/1.23/foo.box'
        assert box_version == '1.23'


@patch('mech.utils.tar_cmd')
def test_add_box_file_do_not_save(mock_tar_cmd):
    """Test add_box_file."""
    mock_tarfile_open = MagicMock()
    mock_tarfile_open.returncode = 0
    mock_tarfile_open.return_value.getnames.return_value = ['aaa', 'some.vmx']
    mock_tar_cmd.return_value = None
    with patch('tarfile.open', mock_tarfile_open, create=True):
        box, box_version = mech.utils.add_box_file(box='bento/ubuntu',
                                                   box_version='1.23',
                                                   filename='/tmp/foo.box',
                                                   save=False)
        mock_tarfile_open.assert_called()
        mock_tar_cmd.assert_called()
        assert box == '/tmp/foo.box'
        assert box_version == '1.23'


@patch('mech.utils.tar_cmd')
def test_add_box_file_that_has_leading_slashes(mock_tar_cmd):
    """Test add_box_file."""
    mock_tarfile_open = MagicMock()
    mock_tarfile_open.returncode = 0
    mock_tarfile_open.return_value.getnames.return_value = ['/aaa', '/some.vmx']
    mock_tar_cmd.return_value = None
    with raises(SystemExit, match=r"This box is comprised of filenames starting with"):
        with patch('tarfile.open', mock_tarfile_open, create=True):
            mech.utils.add_box_file(box='bento/ubuntu', box_version='1.23', filename='/tmp/foo.box')


def test_add_mechfile_with_empty_mechfile():
    """Test add_mechfile."""
    mech.utils.add_mechfile(mechfile_entry={})


@patch('mech.utils.add_box_file', return_value='')
def test_add_mechfile_with_boxfile(mock_add_box_file, mechfile_one_entry_with_file):
    """Test add_mechfile."""
    mech.utils.add_mechfile(mechfile_entry=mechfile_one_entry_with_file)
    mock_add_box_file.assert_called()


@patch('requests.get')
@patch('mech.utils.locate')
def test_add_box_url(mock_locate, mock_requests_get, catalog_as_json):
    """Test init_box."""
    mock_locate.return_value = False
    mock_requests_get.return_value.status_code = 200
    mock_requests_get.return_value.json.return_value = catalog_as_json
    got = mech.utils.add_box_url(name='first', box='abox', box_version='aver', url='')
    assert got is None


@patch('os.getlogin', return_value='bob')
@patch('os.path.expanduser', return_value='/home/bob/id_rsa.pub')
def test_get_info_for_auth(mock_path_expanduser, mock_getlogin):
    """Test get_info_for_auth."""
    expected = {'auth': {'username': 'bob', 'pub_key': '/home/bob/id_rsa.pub', 'mech_use': False}}
    got = mech.utils.get_info_for_auth()
    assert got == expected


@patch('sys.platform', return_value='atari')
def test_get_provider(mock_sys_platform):
    """Test get_provider simulating an Atari 800XL."""
    a_mock = MagicMock()
    another_mock = MagicMock()
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    another_mock.communicate = yet_another_mock
    another_mock.communicate.returncode = 0
    another_mock.returncode = 0
    a_mock.return_value = another_mock
    a_mock.returncode = 0
    with patch('subprocess.Popen', a_mock):
        provider = mech.utils.get_provider('/tmp/vmrun')
        assert provider == 'ws'


@patch('sys.platform', return_value='nt')
def test_get_provider_simulate_win(mock_sys_platform):
    """Test get_provider simulating windows."""
    a_mock = MagicMock()
    another_mock = MagicMock()
    yet_another_mock = MagicMock()
    yet_another_mock.returncode = 0
    another_mock.communicate = yet_another_mock
    another_mock.communicate.returncode = 0
    another_mock.returncode = 0
    a_mock.return_value = another_mock
    a_mock.returncode = 0
    with patch('subprocess.Popen', a_mock):
        provider = mech.utils.get_provider('/tmp/vmrun')
        assert provider == 'ws'


def test_valid_provider():
    """Test valid_provider."""
    assert mech.utils.valid_provider('vmware')
    assert mech.utils.valid_provider('virtualbox')
    assert not mech.utils.valid_provider(None)
    assert not mech.utils.valid_provider('')
    assert not mech.utils.valid_provider('atari')


def test_vm_ready_based_on_state():
    """Test vm_ready_based_on_state."""
    assert mech.utils.vm_ready_based_on_state('powered on')
    assert mech.utils.vm_ready_based_on_state('running')
    assert mech.utils.vm_ready_based_on_state('started')
    assert not mech.utils.vm_ready_based_on_state('unkown')
    assert not mech.utils.vm_ready_based_on_state('paused')
    assert not mech.utils.vm_ready_based_on_state('power off')



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: auth tests"""
import re
import subprocess

import pytest


@pytest.mark.int
def test_int_auth(helpers):
    """Auth testing."""

    test_dir = "tests/int/auth"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    # "up" with "add-me" and "use-me" options
    command = "mech init -a -u bento/ubuntu-18.04"
    expected_lines = [r"init"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # "list" detailed view shows psk
    command = "mech ls -d"
    expected_lines = [r"id_rsa.pub"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # "up" with 'remove-vagrant' option
    command = "mech up -r"
    expected_lines = [r"started", "Removing"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # make sure we can run a command using our psk (but do not
    # add entry to our known hosts file)
    command = """first_ip=`mech ip first`;
         ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null \
-o LogLevel=QUIET ${first_ip} -C uptime"""
    expected_lines = [r"load average"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up at the end
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: simple ones (including smoke)"""
import re
import subprocess
import platform

import pytest


@pytest.mark.int
def test_int_no_args():
    """Test without any args"""
    return_value, out = subprocess.getstatusoutput('mech')
    assert re.match(r'Usage: mech ', out)
    assert return_value == 1


@pytest.mark.int
def test_int_version():
    """Test '--version'."""
    return_value, out = subprocess.getstatusoutput('mech --version')
    assert re.match(r'mech v[0-9]+\.[0-9]+\.[0-9]', out)
    assert return_value == 0


@pytest.mark.int
def test_int_help():
    """Test '--help'."""
    return_value, out = subprocess.getstatusoutput('mech --help')
    assert re.match(r'Usage: mech ', out)
    assert return_value == 0


@pytest.mark.int
def test_int_no_mechfile(helpers):
    """Test when no Mechfile."""
    test_dir = "tests/int/no_mechfile"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)
    command = "mech ls"
    return_value, out = subprocess.getstatusoutput(command)
    assert re.search(r'Could not find a Mechfile', out)
    assert return_value == 1


@pytest.mark.int
def test_int_smoke(helpers):
    """Smoke test most options."""

    test_dir = "tests/int/simple"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    # ensure we need to provide more args
    commands = ["mech box", "mech init", "mech ip", "mech ps", "mech scp", "mech snapshot",
                "mech snapshot save", "mech snapshot save snap1", "mech snapshot delete",
                "mech snapshot remove", "mech ssh"]
    expected = "Usage: mech "
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stdout == ''
        assert results.returncode == 1
        assert re.search(expected, stderr)

    # should init
    command = "mech init mrlesmithjr/alpine311"
    expected_lines = ["Initializing", "Loading metadata", "has been initialized", "mech up"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should start
    command = "mech up"
    expected_lines = ["could not be found", "vmware_desktop",
                      "integrity", "Extracting", "Added network",
                      "Bringing machine", "Getting IP", "Sharing folders",
                      "started", "Provisioning"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure we get proper response from bad instance name
    commands = ["mech ip first2"]
    expected = "was not found in the Mechfile"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stdout == ''
        assert results.returncode == 1
        assert re.search(expected, stderr)

    # should be able to re-up, verify 'start' alias works, too
    commands = ["mech up", "mech start"]
    expected_lines = ["Bringing machine", "Getting IP", "Sharing folders",
                      "started", "Provisioning"]
    for command in commands:
        results = subprocess.run(commands, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech ps'
    command = "mech ps first"
    expected_lines = ["/sbin/init"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech global-status'
    command = "mech global-status"
    expected_lines = [test_dir + '/.mech/first/']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech list'
    commands = ["mech ls", "mech list"]
    expected_lines = ['first', 'alpine']
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech stop'
    command = "mech stop"
    expected_lines = ['Stopped']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech stop' again
    command = "mech stop"
    expected_lines = ['Not stopped']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert re.search('The virtual machine is not powered on', stderr, re.MULTILINE)
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech start'
    command = "mech start"
    expected_lines = ['started']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech pause'
    command = "mech pause"
    expected_lines = ['Paused']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech resume'
    command = "mech resume"
    expected_lines = ['resumed']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech suspend'
    command = "mech suspend"
    expected_lines = ['Suspended']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech resume' after suspend
    command = "mech resume"
    expected_lines = ['started']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech ssh' (different forms)
    commands = ["mech ssh -c 'uptime' first", "mech ssh --command 'uptime' first"]
    expected_lines = ['load average']
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech scp' to guest
    command = "date > now; mech scp now first:/tmp"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # test 'mech scp' from guest
    command = "mech scp first:/tmp/now ."
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # test 'mech ip first'
    command = "mech ip first"
    expected = r"[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    if platform.system() == "Linux":
        # test "mech port"
        command = "mech port"
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stdout == ''
        assert stderr.strip() == 'This command is not supported on this OS.'
        assert results.returncode == 1
    else:
        # test "mech port"
        command = "mech port"
        expected = "Total port forwardings: 0"
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech box list" (and alias)
    commands = ["mech box list", "mech box ls"]
    expected = r"alpine"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot list" (and alias)
    commands = ["mech snapshot list", "mech snapshot ls"]
    expected = "Total snapshots: 0"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot save"
    command = "mech snapshot save snap1 first"
    expected = "taken"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # test "mech snapshot save" with same args again
    command = "mech snapshot save snap1 first"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert re.search('A snapshot with the name already exists', stderr)
    assert results.returncode == 1

    # test "mech snapshot list" (and alias) again (now that we have one)
    commands = ["mech snapshot list", "mech snapshot ls"]
    expected = "Total snapshots: 1"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot delete"
    command = "mech snapshot delete snap1 first"
    expected = "deleted"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up at the end
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# Copyright (c) 2020 Mike Kinney

"""mech tests"""
from unittest.mock import patch
from pytest import raises

import mech.command
import mech.mech
import mech.vmrun
import mech.mech_instance


def test_mech_instance_name_not_being_in_instance():
    """Test mech instance class."""
    with raises(SystemExit):
        mech.mech.MechInstance('first', {})


@patch('mech.vmrun.VMrun.get_guest_ip_address', return_value="192.168.1.100")
@patch('mech.utils.locate', return_value='/tmp/first/some.vmx')
def test_mech_instance(mock_locate, mock_get_ip_address,
                       mechfile_one_entry_with_auth_and_mech_use):
    """Test mech instance class."""
    inst = mech.mech.MechInstance('first', mechfile_one_entry_with_auth_and_mech_use)
    mock_locate.assert_called()
    print('inst.ip:{}'.format(inst.ip))
    inst.config_ssh()
    mock_get_ip_address.assert_called()
